from pumapy.utilities.workspace import Workspace
from pumapy.utilities.timer import Timer
import pumapy.utilities.generic_checks as check
import numpy as np


def generate_sphere(shape, center, diameter, segmented=True):
    """ Generation of a sphere at a given point and diameter

        :param shape: size of 3D domain (x,y,z)
        :type shape: (int, int, int)
        :param center: centerpoint of sphere (x,y,z)
        :type center: (int, int, int)
        :param diameter: diameter of the random spheres in voxels
        :type diameter: float
        :param segmented: return a domain that is already segmented (i.e. each sphere with unique ID) or
            with grayscales 0-255 with threshold=128 for the input diameter
        :type segmented: bool
        :return: domain with sphere with input diameter
        :rtype: Workspace

        :Example:
        >>> import pumapy as puma
        >>> ws = puma.generate_sphere((100, 100, 100), (50, 50, 50), 80)
        >>> puma.render_volume(ws.matrix[:ws.matrix.shape[0]//2])
    """

    generator = GeneratorSphere(shape, center, diameter)

    generator.error_check()

    generator.log_input()
    ws = generator.generate()
    generator.log_output()

    if segmented:
        ws.binarize(128)
    return ws


class GeneratorSphere:
    def __init__(self, size, center, diameter):
        self._workspace = Workspace()
        self._size = size
        self._center = center
        self._diameter = diameter

        self._size_padded = (self._size[0]+diameter, self._size[1]+diameter, self._size[2]+diameter)
        self._start = int(0.5 * diameter)
        self._endX = size[0] + self._start
        self._endY = size[1] + self._start
        self._endZ = size[2] + self._start

        # size of the created sphere block
        self._spheresize = int(self._diameter+4)
        if self._spheresize % 2 == 0:
            self._size_m = self._spheresize / 2
            self._size_p = self._spheresize / 2
        else:
            self._size_m = int(self._spheresize / 2.)+1
            self._size_p = int(self._spheresize / 2.)

    def generate(self):
        t = Timer()
        self._workspace.resize_new_matrix(self._size_padded)
        self._generate_sphere()
        print("Generated in: " + str(t.elapsed()) + " seconds")
        return self._workspace

    def _generate_sphere(self):
        sphere = get_sphere(self._diameter)
        [x,y,z] = self._center + np.array([self._start, self._start, self._start])
        xm = int(x - self._size_m)
        xp = int(x + self._size_p)
        ym = int(y - self._size_m)
        yp = int(y + self._size_p)
        zm = int(z - self._size_m)
        zp = int(z + self._size_p)

        self._workspace.matrix[xm:xp,ym:yp,zm:zp] += sphere[:,:,:]
        self._workspace.matrix = self._workspace.matrix[self._start:self._endX,self._start:self._endY,self._start:self._endZ]

    def log_input(self):
        self._workspace.log.log_section("Generating Sphere")
        self._workspace.log.log_value("Domain Size", self._size)
        self._workspace.log.log_value("Sphere Center", self._center)
        self._workspace.log.log_value("Diameter", self._diameter)
        self._workspace.log.write_log()

    def log_output(self):
        self._workspace.log.log_section("Finished Sphere Generation")
        self._workspace.log.write_log()

    def error_check(self):
        check.size_check(self._size)
        if self._center[0] < 0 or self._center[0] >= self._size[0]:
            raise Exception("Error, invalid center. Must be contained within domain size")
        if self._center[1] < 0 or self._center[1] >= self._size[1]:
            raise Exception("Error, invalid center. Must be contained within domain size")
        if self._center[2] < 0 or self._center[2] >= self._size[2]:
            raise Exception("Error, invalid center. Must be contained within domain size")


def get_sphere(diameter):
    radius = diameter / 2.
    shape = (int(diameter+4), int(diameter+4), int(diameter+4))
    center_point = (diameter+4.) / 2. - 0.5
    point = np.array([center_point, center_point, center_point])
    dist = np.linalg.norm(np.indices(shape)-point[:, None, None, None], axis=0)
    sphere = (255.-127.5*(1+np.tanh((dist-radius))))
    sphere[sphere < 0] = 0
    sphere = sphere.astype(np.uint16)
    return sphere


def get_circle(diameter):
    radius = diameter / 2.
    shape = (int(diameter+4), int(diameter+4))
    center_point = (int(diameter+4)) / 2.
    point = np.array([center_point, center_point])
    dist = np.linalg.norm(np.indices(shape)+0.5-point[:, None, None], axis=0)
    circle = (255.-128*(1+np.tanh(2*(dist-radius))))
    circle[circle < 0] = 0
    circle = circle.astype(np.uint16)
    return circle


def get_circle_odd(diameter):
    radius = diameter / 2.
    if int(diameter) % 2 == 0:
        shape = (int(diameter+5), int(diameter+5))
        center_point = (int(diameter + 5)) / 2.
    else:
        shape = (int(diameter+4), int(diameter+4))
        center_point = (int(diameter + 4)) / 2.
    point = np.array([center_point, center_point])
    dist = np.linalg.norm(np.indices(shape)+0.5-point[:, None, None], axis=0)
    circle = (255.-128*(1+np.tanh(2*(dist-radius))))
    circle[circle < 0] = 0
    circle = circle.astype(np.uint16)
    return circle, shape[0]


def get_circle_even(diameter):
    radius = diameter / 2.
    if int(diameter) % 2 == 0:
        shape = (int(diameter+4), int(diameter+4))
        center_point = (int(diameter + 4)) / 2.
    else:
        shape = (int(diameter+5), int(diameter+5))
        center_point = (int(diameter + 5)) / 2.
    point = np.array([center_point, center_point])
    dist = np.linalg.norm(np.indices(shape)+0.5-point[:, None, None], axis=0)
    circle = (255.-128*(1+np.tanh(2*(dist-radius))))
    circle[circle < 0] = 0
    circle = circle.astype(np.uint16)
    return circle, shape[0]



from pumapy.physicsmodels.mpsa_elasticity import Elasticity
from pumapy.physicsmodels.property_maps import ElasticityMap
from scipy.optimize import fsolve
import numpy as np


def compute_elasticity(workspace, elast_map, direction, side_bc='p', prescribed_bc=None, tolerance=1e-4,
                       maxiter=100000, solver_type='gmres', display_iter=True, print_matrices=(0, 0, 0, 0, 0)):
    """ Compute the effective elasticity coefficient
        (N.B. 0 material ID in workspace refers to air unless otherwise specified)

        :param workspace: domain
        :type workspace: pumapy.Workspace
        :param elast_map: local elasticity of the constituents
        :type elast_map: pumapy.ElasticityMap
        :param direction: direction for solve ('x','y', 'z', 'yz', 'xz', 'xy')
        :type direction: string
        :param side_bc: side boundary conditions can be symmetric ('s'), periodic ('p'), dirichlet ('d') or free ('f')
        :type side_bc: string
        :param prescribed_bc: 3D array holding dirichlet BC
        :type prescribed_bc: pumapy.ElasticityBC or None
        :param tolerance: tolerance for iterative solver
        :type: tolerance: float
        :param maxiter: maximum Iterations for solver
        :type maxiter: int
        :param solver_type: solver type, options: 'bicgstab', 'cg', 'gmres' (default), 'direct'
        :type solver_type: string
        :param display_iter: display iterations and residual
        :type display_iter: bool
        :param print_matrices: corresponding to b, E, A, u, s decimal places. If 0, they are not printed
        :type print_matrices: (int, int, int, int, int)
        :return: elasticity, displacement field, direct stresses (sigma xx, yy, zz), shear stresses (tau yz, xz, xy)
        :rtype: ((float, float, float, float, float, float), numpy.ndarray, numpy.ndarray, numpy.ndarray)

        :Example:
        >>> import pumapy as puma
        >>> ws = puma.Workspace.from_shape_value((20, 20, 20), 1)
        >>> ws[int(ws.matrix.shape[0] / 2):] = 2
        >>> elast_map = puma.ElasticityMap()
        >>> elast_map.add_isotropic_material((1, 1), 200, 0.3)
        >>> elast_map.add_isotropic_material((2, 2), 400, 0.1)
        >>> C, u, s, t = puma.compute_elasticity(ws, elast_map, direction='x', side_bc='f', solver_type="direct")
        >>> print(C)
    """
    if isinstance(elast_map, ElasticityMap):
        solver = Elasticity(workspace, elast_map, direction, side_bc, prescribed_bc, tolerance, maxiter,
                            solver_type, display_iter, print_matrices)
    else:
        raise Exception("elast_map has to be an ElasticityMap")

    solver.error_check()

    solver.log_input()
    solver.compute()
    solver.log_output()
    return solver.Ceff, solver.u, solver.s, solver.t


def compute_stress_analysis(workspace, elast_map, prescribed_bc, side_bc='p', tolerance=1e-4,
                            maxiter=100000, solver_type='gmres', display_iter=True, print_matrices=(0, 0, 0, 0, 0)):
    """ Compute stress analysis
        (N.B. 0 material ID in workspace refers to air unless otherwise specified)

        :param workspace: domain
        :type workspace: pumapy.Workspace
        :param elast_map: local elasticity of the constituents
        :type elast_map: pumapy.ElasticityMap
        :param prescribed_bc: 3D array holding dirichlet BC
        :type prescribed_bc: pumapy.ElasticityBC
        :param side_bc: side boundary conditions can be symmetric ('s'), periodic ('p'), dirichlet ('d') or free ('f')
        :type side_bc: string
        :param tolerance: tolerance for iterative solver
        :type tolerance: float
        :param maxiter: maximum Iterations for solver
        :type maxiter: int
        :param solver_type: solver type, options: 'bicgstab', 'cg', 'gmres' (default), 'direct'
        :type solver_type: string
        :param display_iter: display iterations and residual
        :type display_iter: bool
        :param print_matrices: corresponding to b, E, A, u, s decimal places. If 0, they are not printed
        :type print_matrices: (int, int, int, int, int)
        :return: displacement field, direct stresses (sigma xx, yy, zz), shear stresses (tau yz, xz, xy)
        :rtype: (numpy.ndarray, numpy.ndarray, numpy.ndarray)

        :Example:
        >>> import pumapy as puma
        >>> ws = puma.Workspace.from_shape_value((20, 20, 20), 1)
        >>> ws[ws.matrix.shape[0]//2:] = 2
        >>> elast_map = puma.ElasticityMap()
        >>> elast_map.add_isotropic_material((1, 1), 200, 0.3)
        >>> elast_map.add_isotropic_material((2, 2), 400, 0.1)
        >>> bc = puma.ElasticityBC(ws)
        >>> bc.dirichlet[0] = 0  # hold x -ve face
        >>> bc.dirichlet[-1, :, :, 0] = 1   # displace x +ve face by 1 in x direction
        >>> bc.dirichlet[-1, :, :, 1:] = 0  # hold x +ve face in y and z directions
        >>> u, s, t = puma.compute_stress_analysis(ws, elast_map, bc, side_bc='f', solver_type="direct")
        >>> puma.render_volume(u[:, :, :, 1], cmap='jet')  # displacement magnitude in y direction
    """
    if isinstance(elast_map, ElasticityMap):
        solver = Elasticity(workspace, elast_map, None, side_bc, prescribed_bc, tolerance, maxiter,
                            solver_type, display_iter, print_matrices)
    else:
        raise Exception("elast_map has to be an ElasticityMap")

    solver.error_check()

    solver.log_input()
    solver.compute()
    solver.log_output()
    return solver.u, solver.s, solver.t


def get_E_nu_from_elasticity(C11, C12, C13, C22, C23, C33):
    """ Compute Young's moduli E1, E2, E3 and Poisson's ratios nu12, nu23, nu31 from symmetric elastic stiffness tensor
        :param C11: elasticity tensor component
        :type C11: float
        :param C12: elasticity tensor component
        :type C12: float
        :param C13: elasticity tensor component
        :type C13: float
        :param C22: elasticity tensor component
        :type C22: float
        :param C23: elasticity tensor component
        :type C23: float
        :param C33: elasticity tensor component
        :type C33: float
        :return: Young's moduli E1, E2, E3 and Poisson's ratios nu12, nu23, nu31
        :rtype: (float, float, float, float, float, float)
    """
    def eqs(unknowns, *cs):
        E1, E2, E3, nu12, nu23, nu31 = unknowns
        c11, c12, c31, c22, c23, c33 = cs
        nu21 = nu12 * E2 / E1
        nu32 = nu23 * E3 / E2
        nu13 = nu31 * E1 / E3
        c0 = (1 - nu12 * nu21 - nu23 * nu32 - nu13 * nu31 - 2 * nu21 * nu32 * nu13) / (E1 * E2 * E3)
        return (c11 - (1 - nu23 * nu32) / (c0 * E2 * E3),
                c22 - (1 - nu13 * nu31) / (c0 * E1 * E3),
                c33 - (1 - nu21 * nu12) / (c0 * E2 * E1),
                c12 - (nu12 + nu32 * nu13) / (c0 * E1 * E3),
                c23 - (nu23 + nu21 * nu13) / (c0 * E1 * E2),
                c31 - (nu31 + nu21 * nu32) / (c0 * E2 * E3))

    s = fsolve(eqs, np.full(6, 0.1), args=(C11, C12, C13, C22, C23, C33))
    [print(i, j) for i, j in zip(["E1", "E2", "E3", "nu12", "nu23", "nu31"], s)]
    return s



import unittest
import pumapy as puma
import numpy as np


# the following works locally, but not on github workflow
# import multiprocessing
#
# def test_plot_slices():
#     ws = puma.import_3Dtiff(puma.path_to_example_file("100_fiberform.tif"), 1.3e-6)
#     puma.plot_slices(ws)
#
# def test_compare_slices():
#     ws = puma.import_3Dtiff(puma.path_to_example_file("100_fiberform.tif"), 1.3e-6)
#     ws2 = ws.copy()
#     ws2.binarize_range((100, 255))
#     puma.compare_slices(ws, ws2)
#
# def run_test(self_test, function):
#     p = multiprocessing.Process(target=function)
#     p.start()
#     p.join(3)
#
#     if p.is_alive():
#         print("Function executed for 3 seconds with no errors, this is a planned timeout.")
#         p.terminate()
#         p.join()
#     else:
#         print("Exception raised in detached process.")
#         # self_test.assertEqual(1, 0)
#
# class TestSlicer(unittest.TestCase):
#
#     def test_plot_slices(self):
#         run_test(self, test_plot_slices)
#
#     def test_compare_slices(self):
#         run_test(self, test_compare_slices)


class TestRender(unittest.TestCase):

    def test_render_volume(self):
        ws = puma.Workspace.from_shape_value((5, 5, 5), 1)

        # turn to True to visually inspect plots
        plot = False

        # trying varying different options
        puma.render_volume(ws, cutoff=None, solid_color=(1,1,1), style='surface', origin=(0., 0., 0.),
                           window_size=(1920, 1200), opacity=1., background=(0.3, 0.3, 0.3), show_grid=True, plot_directly=plot,
                           show_axes=True, show_outline=True, cmap='gray', add_to_plot=None, notebook=False)
        puma.render_volume(ws, cutoff=None, solid_color=(1,1,1), style='edges', origin=(0., 0., 0.),
                           window_size=(1920, 1200), opacity=1., background=(0.3, 0.3, 0.3), show_grid=True, plot_directly=plot,
                           show_axes=True, show_outline=True, cmap='gray', add_to_plot=None, notebook=False)
        puma.render_volume(ws, cutoff=None, solid_color=(1,1,1), style='wireframe', origin=(0., 0., 0.),
                           window_size=(1920, 1200), opacity=1., background=(0.3, 0.3, 0.3), show_grid=True, plot_directly=plot,
                           show_axes=True, show_outline=True, cmap='gray', add_to_plot=None, notebook=False)
        puma.render_volume(ws, cutoff=None, solid_color=(1,1,1), style='points', origin=(0., 0., 0.),
                           window_size=(1920, 1200), opacity=1., background=(0.3, 0.3, 0.3), show_grid=True, plot_directly=plot,
                           show_axes=True, show_outline=True, cmap='gray', add_to_plot=None, notebook=False)
        p = puma.render_volume(ws, cutoff=None, solid_color=(1,1,1), style='points', origin=(0., 0., 0.),
                               window_size=(1920, 1200), opacity=1., background=(0.3, 0.3, 0.3), show_grid=True, plot_directly=plot,
                               show_axes=True, show_outline=True, cmap='gray', add_to_plot=None, notebook=False)
        puma.render_volume(ws, cutoff=(1, 1), solid_color=None, style='points', origin=(0., 0., 0.),
                           window_size=(1920, 1200), opacity=1., background=(0.3, 0.3, 0.3), show_grid=False, plot_directly=plot,
                           show_axes=False, show_outline=False, cmap='gray', add_to_plot=p, notebook=False)

    def test_render_contour(self):
        ws = puma.import_3Dtiff(puma.path_to_example_file("100_fiberform.tif"), 1.3e-6)

        # turn to True to visually inspect plots
        plot = False

        # trying varying different options
        puma.render_contour(ws, cutoff=(90, 255), solid_color=(1., 1., 1.), style='surface', origin=(0., 0., 0.),
                            window_size=(1920, 1200), opacity=1., background=(0.3, 0.3, 0.3), show_grid=True, plot_directly=plot,
                            show_axes=True, show_outline=True, add_to_plot=None, notebook=False)
        puma.render_contour(ws, cutoff=(90, 255), solid_color=(1., 1., 1.), style='edges', origin=(0., 0., 0.),
                            window_size=(1920, 1200), opacity=1., background=(0.3, 0.3, 0.3), show_grid=True, plot_directly=plot,
                            show_axes=True, show_outline=True, add_to_plot=None, notebook=False)
        puma.render_contour(ws, cutoff=(90, 255), solid_color=(1., 1., 1.), style='wireframe', origin=(0., 0., 0.),
                            window_size=(1920, 1200), opacity=1., background=(0.3, 0.3, 0.3), show_grid=True, plot_directly=plot,
                            show_axes=True, show_outline=True, add_to_plot=None, notebook=False)
        puma.render_contour(ws, cutoff=(90, 255), solid_color=(1., 1., 1.), style='points', origin=(0., 0., 0.),
                            window_size=(1920, 1200), opacity=1., background=(0.3, 0.3, 0.3), show_grid=True, plot_directly=plot,
                            show_axes=False, show_outline=False, add_to_plot=None, notebook=False)

    def test_render_orientation(self):
        ws = puma.Workspace.from_shape_vector((5, 6, 2), (0.4, 2, 5))

        # turn to True to visually inspect plots
        plot = False

        # trying varying different options
        puma.render_orientation(ws, scale_factor=1., solid_color=None, style='surface', origin=(0., 0., 0.),
                                window_size=(1920, 1200), opacity=1., background=(0.3, 0.3, 0.3), show_grid=True, cmap=None,
                                plot_directly=plot, show_axes=True, show_outline=True, add_to_plot=None, notebook=False,
                                sampling=None)
        puma.render_orientation(ws, scale_factor=0.5, solid_color=None, style='points', origin=(0., 0., 0.),
                                window_size=(1920, 1200), opacity=1., background=(0.3, 0.3, 0.3), show_grid=True, cmap=None,
                                plot_directly=plot, show_axes=True, show_outline=True, add_to_plot=None, notebook=False,
                                sampling=1000)
        puma.render_orientation(ws, scale_factor=0.5, solid_color=None, style='wireframe', origin=(0., 0., 0.),
                                window_size=(1920, 1200), opacity=1., background=(0.3, 0.3, 0.3), show_grid=True, cmap=None,
                                plot_directly=plot, show_axes=True, show_outline=True, add_to_plot=None, notebook=False,
                                sampling=5)
        puma.render_orientation(ws, scale_factor=1., solid_color=None, style='edges', origin=(0., 0., 0.),
                                window_size=(1920, 1200), opacity=1., background=(0.3, 0.3, 0.3), show_grid=True, cmap=None,
                                plot_directly=plot, show_axes=False, show_outline=False, add_to_plot=None, notebook=False,
                                sampling=None)

    def test_render_warp(self):
        ws = puma.Workspace.from_shape_value((5, 6, 2), 1)
        ws.voxel_length = 1
        ws.orientation = np.random.random_sample((5, 6, 2, 3))
        ws.orientation /= ws.orientation_magnitude()[:, :, :, np.newaxis]  # normalize to unit vectors

        # turn to True to visually inspect plots
        plot = False

        # trying varying different options
        puma.render_warp(ws, scale_factor=1., color_by='magnitude', style='surface', origin=(0., 0., 0.),
                         window_size=(1920, 1200), opacity=1., background=(0.3, 0.3, 0.3), show_grid=True, cmap='jet',
                         plot_directly=plot, show_axes=True, show_outline=True, add_to_plot=None, notebook=False)
        puma.render_warp(ws, scale_factor=1., color_by='x', style='points', origin=(0., 0., 0.),
                         window_size=(1920, 1200), opacity=1., background=(0.3, 0.3, 0.3), show_grid=True, cmap='jet',
                         plot_directly=plot, show_axes=False, show_outline=False, add_to_plot=None, notebook=False)
        puma.render_warp(ws, scale_factor=1., color_by='y', style='wireframe', origin=(0., 0., 0.),
                         window_size=(1920, 1200), opacity=1., background=(0.3, 0.3, 0.3), show_grid=True, cmap='jet',
                         plot_directly=plot, show_axes=False, show_outline=False, add_to_plot=None, notebook=False)
        puma.render_warp(ws, scale_factor=1., color_by='z', style='edges', origin=(0., 0., 0.),
                         window_size=(1920, 1200), opacity=1., background=(0.3, 0.3, 0.3), show_grid=True, cmap='jet',
                         plot_directly=plot, show_axes=False, show_outline=False, add_to_plot=None, notebook=False)

    def test_render_contour_multiphase(self):
        ws = puma.import_3Dtiff(puma.path_to_example_file("100_fiberform.tif"), 1.3e-6)

        # turn to True to visually inspect plots
        plot = False

        # trying varying different options
        puma.render_contour_multiphase(ws, cutoffs=((100, 150), (150, 255)), solid_colors=None, style='surface', origin=(0., 0., 0.),
                                       window_size=(1920, 1200), opacity=1., background=(0.3, 0.3, 0.3), show_grid=True,
                                       plot_directly=plot, show_axes=True, show_outline=True, add_to_plot=None, notebook=False)
        puma.render_contour_multiphase(ws, cutoffs=((100, 150), (150, 230), (230, 255)), solid_colors=((1,1,1), (0.6, 0.6, 0.6), (0.3, 0.3, 0.3)),
                                       style='surface', origin=(0., 0., 0.),
                                       window_size=(1920, 1200), opacity=1., background=(0.3, 0.3, 0.3), show_grid=True,
                                       plot_directly=plot, show_axes=False, show_outline=True, add_to_plot=None, notebook=False)
        puma.render_contour_multiphase(ws, cutoffs=((100, 150), (150, 255)), solid_colors=None, style='wireframe', origin=(0., 0., 0.),
                                       window_size=(1920, 1200), opacity=1., background=(0.3, 0.3, 0.3), show_grid=True,
                                       plot_directly=plot, show_axes=True, show_outline=False, add_to_plot=None, notebook=False)
        puma.render_contour_multiphase(ws, cutoffs=((100, 150), (150, 255)), solid_colors=None, style='edges', origin=(0., 0., 0.),
                                       window_size=(1920, 1200), opacity=1., background=(0.3, 0.3, 0.3), show_grid=False,
                                       plot_directly=plot, show_axes=True, show_outline=True, add_to_plot=None, notebook=False)


if __name__ == '__main__':
    unittest.main()



from pumapy.materialproperties.volumefraction import compute_volume_fraction
from pumapy.physicsmodels.property_maps import IsotropicConductivityMap
from pumapy.physicsmodels.isotropic_conductivity import IsotropicConductivity


def compute_continuum_tortuosity(workspace, cutoff, direction, side_bc='p', prescribed_bc=None,
                                 tolerance=1e-4, maxiter=10000, solver_type='cg', display_iter=True):
    """ Compute the tortuosity modelling the local conductivity as isotropic

        :param workspace: domain
        :type workspace: pumapy.Workspace
        :param cutoff: to binarize domain
        :type cutoff: (int, int)
        :param direction: direction for solve ('x','y', or 'z')
        :type direction: string
        :param side_bc: side boundary conditions (string) can be symmetric ('s'), periodic ('p') or dirichlet ('d')
        :type side_bc: string
        :param prescribed_bc: 3D array holding dirichlet BC
        :type prescribed_bc: pumapy.ConductivityBC or None
        :param tolerance: tolerance for iterative solver
        :type tolerance: float
        :param maxiter: maximum Iterations for solver
        :type maxiter: int
        :param solver_type: solver type, options: 'cg' (default), 'bicgstab', 'direct'
        :type solver_type: string
        :param display_iter: display iterations and residual
        :type display_iter: bool
        :return: tortuosity, diffusivity, porosity, concentration field
        :rtype: ((float, float, float), float, float, numpy.ndarray)
        :Example:
        >>> import pumapy as puma
        >>> ws_fiberform = puma.import_3Dtiff(puma.path_to_example_file("200_fiberform.tif"), 1.3e-6) # import image
        >>> n_eff_x, Deff_x, poro, C_x = puma.compute_continuum_tortuosity(ws_fiberform, (0,89), 'x', side_bc='s', tolerance=1e-2, solver_type='cg')
        >>> print('Effective tortuosity factors:', n_eff_x)
    """

    cond_map = IsotropicConductivityMap()
    cond_map.add_material(cutoff, 1)

    if cutoff[0] > 0:
        cond_map.add_material((0, cutoff[0]-1),0)
    cond_map.add_material((cutoff[1]+1,32000),0)

    solver = IsotropicConductivity(workspace, cond_map, direction, side_bc, prescribed_bc,
                                   tolerance, maxiter, solver_type, display_iter)

    solver.error_check()

    solver.log_input()
    solver.compute()
    solver.log_output()

    porosity = compute_volume_fraction(workspace, cutoff)
    eta = [porosity / solver.keff[0], porosity / solver.keff[1], porosity / solver.keff[2]]

    return eta, solver.keff, porosity, solver.T



from pumapy.utilities.workspace import Workspace
import math
import numpy as np


def size_check(size):
    if not (isinstance(size, tuple) and len(size) == 3):
        raise Exception("Error, invalid size, must be tuple of size 3")
    if size[0] <= 0:
        raise Exception("Error, invalid size, must be >= 0")
    if size[1] <= 0:
        raise Exception("Error, invalid size, must be >= 0")
    if size[2] <= 0:
        raise Exception("Error, invalid size, must be >= 0")


def range_inc(var, var_range, var_name):
    if var < var_range[0] or var > var_range[1]:
        raise Exception("Error, invalid " + str(var_name) + ", must be between " + str(var_range[0]) + " and " + str(
            var_range[1]) + ", inclusive")


def range_exc(var, var_range, var_name):
    if var <= var_range[0] or var >= var_range[1]:
        raise Exception("Error, invalid " + str(var_name) + ", must be between " + str(var_range[0]) + " and " + str(
            var_range[1]) + ", exclusive")


def greater_than_inc(var, lim, var_name):
    if var < lim:
        raise Exception("Error, invalid " + str(var_name) + ", must be greater than " + str(lim) + ", inclusive")


def greater_than_exc(var, lim, var_name):
    if var <= lim:
        raise Exception("Error, invalid " + str(var_name) + ", must be greater than " + str(lim) + ", exclusive")


def less_than_inc(var, lim, var_name):
    if var > lim:
        raise Exception("Error, invalid " + str(var_name) + ", must be less than " + str(lim) + ", inclusive")


def less_than_exc(var, lim, var_name):
    if var >= lim:
        raise Exception("Error, invalid " + str(var_name) + ", must be less than " + str(lim) + ", exclusive")


def check_ws_cutoff(workspace, cutoff):
    if not isinstance(workspace, Workspace):
        raise Exception("Need to pass pumapy.Workspace.")
    if workspace.matrix.shape == [1, 1, 1]:
        raise Exception("Empty workspace")
    if len(cutoff) != 2:
        raise Exception("Cutoff should be of length 2. Ex: (128,255)")
    if min(cutoff[0], cutoff[1]) < 0:
        raise Exception("Invalid cutoff. Must be positive")
    if cutoff[0] > cutoff[1]:
        raise Exception("Invalid cutoff. cutoff[0] should be <= cutoff[1]")


def estimate_max_memory(material_property, workspace_shape, solver_type='iterative', need_to_orient=False,
                        permeability_fluid_vf=1.):
    """ Compute an estimate of the extra maximum memory required to run a specified material property on a domain

        :param material_property: property to estimate, options:
        'anisotropic_conductivity', 'isotropic_conductivity', 'tortuosity', 'elasticity', 'permeability'
        :type material_property: string
        :param workspace_shape: size of the domain to compute
        :type workspace_shape: (int, int, int)
        :param solver_type: type of solver, options: 'direct', 'iterative'
        :type solver_type: string
        :param need_to_orient: domain with orientation (needed for anisotropic conductivity and elasticity)
        :type need_to_orient: bool
        :param permeability_fluid_vf: volume fraction of the fluid (needed for permeability)
        :type permeability_fluid_vf: float
        :return: number of Bytes
        :rtype: int
    """

    # missing properties: 'orientation', 'radiation'
    mat_properties = ['anisotropic_conductivity', 'isotropic_conductivity', 'tortuosity', 'elasticity', 'permeability']

    if material_property not in mat_properties:
        raise Exception(f"material_property input can only be one of the following types: {mat_properties}")

    def convert_bytes_size(size_bytes):
        if size_bytes == 0:
            return "0B"
        size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
        i = int(math.floor(math.log(size_bytes, 1024)))
        p = math.pow(1024, i)
        s = round(size_bytes / p, 2)
        return "%s %s" % (s, size_name[i])

    # sizes in bytes
    float_size = np.zeros(1, dtype=float).itemsize
    uint32_size = np.zeros(1, dtype=np.uint32).itemsize
    uint16_size = np.zeros(1, dtype=np.uint16).itemsize

    len_x, len_y, len_z = workspace_shape
    total_bytes = 0

    # General idea: peak memory should be reached at Amat sparse matrix creation
    if material_property in ["anisotropic_conductivity", "elasticity"]:

        # because of the use of padded domains
        len_x += 2
        len_y += 2
        len_z += 2

        if material_property == "anisotropic_conductivity":
            values_in_Amat_rows = 27
            dof = 1.  # degrees of freedom
        else:
            values_in_Amat_rows = 81
            dof = 3.

        len_xyz = len_x * len_y * len_z

        # Amat size (V, I, J)
        A_rows = dof * values_in_Amat_rows * len_xyz
        total_bytes += 2 * (A_rows * float_size + 2 * A_rows * uint32_size)

        # bvec size
        bvec = dof * len_y * len_z * float_size + 2 * len_xyz * uint32_size
        total_bytes += bvec

        # ws_pad_size
        total_bytes += len_xyz * uint16_size

        # case dependent variables
        if solver_type != "direct":
            total_bytes += dof * len_xyz * float_size  # initial guess size

        if need_to_orient:
            total_bytes += 3 * len_xyz * float_size  # orient_pad_size

    elif material_property in ["isotropic_conductivity", "tortuosity"]:

        # Amat size (V, I, J)
        A_rows = (len_x - 2) * len_y * len_z * 7 + 2 * len_y * len_z
        total_bytes += 2 * (A_rows * float_size + 2 * A_rows * uint32_size)

        len_xyz = len_x * len_y * len_z

        # bvec size
        total_bytes += len_xyz * float_size

        # ws_pad_size
        total_bytes += len_xyz * uint16_size

        if solver_type != "direct":
            total_bytes += len_xyz * float_size  # initial guess size

    elif material_property == "permeability":

        # Amat size (V, I, J)
        ndofs = permeability_fluid_vf * len_x * len_y * len_z
        A_rows = ndofs * 24 * 24 + ndofs * 24 * 8 + ndofs * 192 + ndofs * 8 * 8
        total_bytes += 2 * (A_rows * float_size + 2 * A_rows * uint32_size)

        # bvec
        total_bytes += 3 * (4 * ndofs * float_size + 2 * 4 * ndofs * uint32_size)

        # mgdlF and resolveF
        total_bytes += 32 * ndofs
        total_bytes += 3.5 * ndofs  # rough

    # elif material_property == "orientation":
    #     pass
    #
    # elif material_property == "radiation":
    #     pass

    print(f"Memory requirement for conductivity simulation: {convert_bytes_size(total_bytes)}")
    return total_bytes


def set_random_seed(seed):
    """ Set random seed for scipy and numpy to make results reproducible.
        NB: if you want to generate the same material twice in the same process, you need to call it twice (see example)

        :param seed: random seed
        :type seed: int

        :Example:
        >>> import pumapy as puma
        >>> puma.set_random_seed(1)
        >>> ws = puma.generate_random_spheres((100, 100, 100), 20, 0.5, allow_intersect=True, segmented=False)
        >>> puma.render_volume(ws)
        >>> puma.set_random_seed(1)  # need to call it again to get the same domain!
        >>> ws = puma.generate_random_spheres((100, 100, 100), 20, 0.5, allow_intersect=True, segmented=False)
        >>> puma.render_volume(ws)
    """
    np.random.seed(seed)



"""
Functions to relying on Connected Component Labeling (CCL)
"""
import numpy as np
from skimage import measure
from pumapy.utilities.generic_checks import check_ws_cutoff
from pumapy.utilities.workspace import Workspace


def identify_porespace(workspace, solid_cutoff, connectivity=None):
    """ Identify the porespace

        :param workspace: domain
        :type workspace: pumapy.Workspace
        :param solid_cutoff: specify the solid range to discard from pores identification
        :type solid_cutoff: (int, int)
        :param connectivity: Maximum number of orthogonal hops to consider a pixel/voxel as a neighbor
        (None automatically gives full connectivity of ``input.ndim``, which for 3D domain is 3, for 2D is 2)
        :type connectivity: int or None
        :return: porespace marked as: 0 solid, 1 largest pore (likely open porosity), >1 other pores
        :rtype: np.ndarray

        :Example:
        >>> import pumapy as puma
        >>> ws = puma.generate_sphere((100, 100, 100), (50, 50, 50), 40, segmented=False)
        >>> ws.binarize_range((1, 253))
        >>> puma.render_volume(ws[:ws.matrix.shape[0] // 2], cutoff=(0, 255))
        >>> pores = puma.identify_porespace(ws, (1, 1))
        >>> puma.render_volume(pores[:pores.shape[0] // 2], cutoff=(0, pores.max()))
    """

    # error check
    check_ws_cutoff(workspace, solid_cutoff)
    ws = workspace.copy()

    # logging
    workspace.log.log_section("Identifying Porespace")
    workspace.log.log_value("Solid Cutoff: ", solid_cutoff)
    workspace.log.write_log()

    ws.binarize_range(solid_cutoff)

    # padding with solid to detect pores along the borders
    array = np.ones((ws.matrix.shape[0] + 2, ws.matrix.shape[1] + 2, ws.matrix.shape[2] + 2))
    array[1:-1, 1:-1, 1:-1] = ws.matrix

    pore_labels = measure.label(array, background=1, connectivity=connectivity)
    pore_labels = pore_labels[1:-1, 1:-1, 1:-1]
    unique_pore_ids, unique_id_counts = np.unique(pore_labels[pore_labels != 0], return_counts=True)
    sorted_unique_pore_ids = unique_pore_ids[np.argsort(unique_id_counts)[::-1]]
    keyarray = np.arange(np.max(pore_labels) + 1)
    keyarray[sorted_unique_pore_ids] = unique_pore_ids
    return keyarray[pore_labels]


def fill_closed_pores(workspace, solid_cutoff, fill_value, return_pores=False):
    """ Identify the porespace and fill closed porosity

        :param workspace: domain
        :type workspace: pumapy.Workspace
        :param solid_cutoff: specify the solid range to discard from pores identification
        :type solid_cutoff: (int, int)
        :param fill_value: value to fill closed porosity with
        :type fill_value: int
        :param return_pores: specifies whether to return identified pores
        :type return_pores: bool
        :return: filled workspace with IDs set as: 0 for void, 1 for solid, fill_value for 
            added filler material. In addition, if return_pores==True, then it also returns 
            the porespace marked as: 0 solid, 1 largest pore (likely open porosity), >1 other pores
        :rtype: pumapy.Workspace

        :Example:
        >>> import pumapy as puma
        >>> ws = puma.generate_random_spheres((100, 100, 100), diameter=20, porosity=0.8, allow_intersect=True, segmented=False)
        >>> puma.render_volume(ws[:ws.matrix.shape[0] // 2])
        >>> ws.binarize_range((1, 250))
        >>> puma.render_volume(ws[:ws.matrix.shape[0] // 2])
        >>> filled_ws, pores = puma.fill_closed_pores(ws, (1, 1), fill_value=2, return_pores=True)
        >>> puma.render_volume(pores, cutoff=(0, pores.max()), cmap='jet')
    """

    pores = identify_porespace(workspace, solid_cutoff)
    if isinstance(pores, bool):
        return False

    ws = workspace.copy()
    ws.binarize_range(solid_cutoff)
    ws[np.where(pores > 1)] = np.uint16(fill_value)

    if return_pores:
        return ws, pores
    else:
        return ws


def remove_rbms(workspace, void_id, direction):
    """ Rigid Body Movements (RBMs) removal in segmented domain

        :param workspace: domain
        :type workspace: pumapy.Workspace
        :param void_id: specify the void ID to discard from RBMs identification
        :type void_id: int
        :param direction: Cartesian direction that has to be connected, options: 'x', 'y', 'z'
        :type direction: str
        :return: workspace without the possible RBMs determined by not being connected from side to side
        :rtype: pumapy.Workspace

        :Example:
        >>> import pumapy as puma
        >>> workspace = puma.import_3Dtiff(puma.path_to_example_file("100_fiberform.tif"))
        >>> new_ws = puma.remove_rbms(workspace, void_cutoff=(0, 103), direction='y')
        >>> puma.render_volume(workspace, (104, 255), solid_color=(1,1,1))
        >>> puma.render_volume(new_ws, (1, new_ws.max()), solid_color=(1,1,1))
    """

    solid = identify_porespace(workspace, (void_id, void_id), connectivity=1)
    uniques = np.unique(solid)
    if uniques[0] == 0:
        uniques = uniques[1:]

    supported_solid = Workspace.from_array(np.full_like(solid, void_id))
    supported_solid.voxel_length = workspace.voxel_length

    # only pass the phases connected from side to side
    for unique in uniques:
        mask = solid == unique
        if direction == 'x':
            if np.any(mask[0]) and np.any(mask[-1]):
                supported_solid[mask] = unique
        elif direction == 'y':
            if np.any(mask[:, 0]) and np.any(mask[:, -1]):
                supported_solid[mask] = unique
        elif direction == 'z':
            if np.any(mask[:, :, 0]) and np.any(mask[:, :, -1]):
                supported_solid[mask] = unique
        else:
            raise Exception("direction input can only be 'x', 'y', 'z'")
    return supported_solid



import pumapy
from pumapy.utilities.logger import Logger, print_warning
import skimage.transform as trans
import numpy as np
from copy import deepcopy
from scipy.ndimage import rotate


class Workspace:
    def __init__(self, **kwargs):
        """ Workspace class holding the domain as a numpy matrix

            :param kwargs: keyword arguments
            - No arguments --> zero matrix of shape (1,1,1)
            - 'shape' argument --> then set all to zeros
            - 'value' --> value set all to full of value
            - 'nparray' --> set Workspace matrix as input array
            - 'scalars' --> bool indicating if a Workspace contains scalars stored in matrix variable
            - 'vectors' --> bool indicating if a Workspace contains vectors stored in orientation variable
            :type kwargs: dict
        """
        self.log = Logger()
        self.voxel_length = 1e-6

        # 3D numpy array holding the geometry (scalar)
        self.matrix = np.zeros((1, 1, 1), dtype=np.uint16)

        # 3D numpy array holding the orientation field (vectors)
        self.orientation = np.zeros((1, 1, 1, 3), dtype=float)

        # setting matrix
        if len(kwargs) > 0:
            if 'shape' in kwargs:
                if isinstance(kwargs['shape'], tuple) and len(kwargs['shape']) == 3:
                    if 'value' in kwargs:
                        self.matrix = np.full(kwargs['shape'], kwargs['value'], dtype=np.uint16)
                    else:
                        self.matrix = np.zeros(kwargs['shape'], dtype=np.uint16)
                else:
                    raise Exception("Wrong shape, tuple with 3 dimensions required.")

            elif 'nparray' in kwargs:
                if isinstance(kwargs['nparray'], np.ndarray):
                    if kwargs['nparray'].ndim == 3:
                        self.matrix = kwargs['nparray'].copy().astype(np.uint16)
                    else:
                        raise Exception("Wrong nparray ndim, 3 dimensions required.")
                else:
                    raise Exception("Wrong nparray type.")
            else:
                raise Exception("Unrecognized keyword.")

        # setting orientation
        if 'vectors' in kwargs and kwargs['vectors']:
            if not isinstance(kwargs['vectors'], bool):
                raise Exception("Orientation input is not a bool.")
            else:
                if 'shape' in kwargs:
                    if isinstance(kwargs['shape'], tuple) and len(kwargs['shape']) == 3:
                        if 'vectorvalue' in kwargs:
                            self.orientation = np.tile(kwargs['vectorvalue'], list(kwargs['shape']) + [1]).astype(float)
                        else:
                            self.orientation = np.zeros(list(kwargs['shape']) + [3], dtype=float)
                    else:
                        raise Exception("Wrong shape, tuple with 3 dimensions required.")
                else:
                    raise Exception("Unrecognized keyword.")

    @classmethod
    def from_shape(cls, shape, orientation=False):
        """ Generate workspace from shape, all matrix value set to zero.

            :param shape: shape of workspace to be created
            :type shape: (int, int, int)
            :param orientation: specify if workspace contains orientation
            :type orientation: bool
            :return: new workspace
            :rtype: pumapy.Workspace

            :Example:
            >>> import pumapy as puma
            >>> ws = puma.Workspace.from_shape((10, 11, 12))
            >>> puma.render_volume(ws, style='edges')
        """
        return cls(shape=shape, vectors=orientation)

    @classmethod
    def from_shape_value(cls, shape, value, orientation=False):
        """ Generate workspace from shape, all matrix values set to the value passed.

            :param shape: shape of workspace to be created
            :type shape: (int, int, int)
            :param value: value to be assigned to the matrix variable
            :type value: int
            :param orientation: specify if workspace contains orientation
            :type orientation: bool
            :return: new workspace
            :rtype: pumapy.Workspace

            :Example:
            >>> import pumapy as puma
            >>> ws = puma.Workspace.from_shape_value((20, 31, 212), 1)
            >>> puma.render_volume(ws, style='edges')
        """
        return cls(shape=shape, value=value, vectors=orientation)

    @classmethod
    def from_shape_vector(cls, shape, vector):
        """ Generate workspace from shape, all orientation vectors set to the vector passed.

            :param shape: shape of workspace to be created
            :type shape: (int, int, int)
            :param vector: vector to be assigned to the orientation variable
            :type vector: (float, float, float)
            :return: new workspace with orientation
            :rtype: pumapy.Workspace

            :Example:
            >>> import pumapy as puma
            >>> ws = puma.Workspace.from_shape_vector((5, 6, 2), (0.4, 2, 5))
            >>> puma.render_orientation(ws)
        """
        return cls(shape=shape, vectorvalue=vector, vectors=True)

    @classmethod
    def from_shape_value_vector(cls, shape, value, vector):
        """ Generate workspace from shape, all matrix and orientation set to the values passed.

            :param shape: shape of workspace to be created
            :type shape: (int, int, int)
            :param value: value to be assigned to the matrix variable
            :type value: int
            :param vector: vector to be assigned to the orientation variable
            :type vector: (float, float, float)
            :return: new workspace with orientation
            :rtype: pumapy.Workspace

            :Example:
            >>> import pumapy as puma
            >>> ws = puma.Workspace.from_shape_value_vector((5, 6, 2), 1, (0.4, 2, 5))
            >>> puma.render_orientation(ws)
        """
        return cls(shape=shape, value=value, vectorvalue=vector, vectors=True)

    @classmethod
    def from_array(cls, nparray):
        """ Generate workspace matrix from numpy array.

            :param nparray: array of shape (X,Y,Z) to be assigned to the matrix variable. NB: array is turned into type uint16
            :type nparray: np.ndarray
            :return: new workspace
            :rtype: pumapy.Workspace

            :Example:
            >>> import pumapy as puma
            >>> array = np.random.randint(5, size=(10, 10, 10))
            >>> ws = puma.Workspace.from_array(array)
            >>> puma.render_volume(ws, style='edges')
        """
        return cls(nparray=nparray, vectors=False)

    def set_voxel_length(self, voxel_length):
        """ Set voxel size, which by default is set to 1e-6

            :param voxel_length: size of a voxel side
            :type voxel_length: float
            :return: None

            :Example:
            >>> import pumapy as puma
            >>> array = np.random.randint(5, size=(10, 10, 10))
            >>> ws = puma.Workspace.from_array(array)
            >>> ws.set_voxel_length(1)  # equivalent to ws.voxel_length = 1
            >>> puma.render_volume(ws, style='edges')
        """
        if not isinstance(voxel_length, int) and not isinstance(voxel_length, float):
            raise Exception("Voxel_length needs to be an int or float, got " + str(type(voxel_length)))
        else:
            self.voxel_length = voxel_length

    def set_matrix(self, nparray):
        """ Set matrix with numpy array

            :param nparray: array of shape (X,Y,Z) to be assigned to the matrix variable
            :type nparray: np.ndarray
            :return: None

            :Example:
            >>> import pumapy as puma
            >>> ws = puma.Workspace()
            >>> nparray = np.random.randint(5, size=(10, 10, 10))
            >>> ws.set_matrix(nparray)  # equivalent to ws.matrix = nparray.copy().astype(np.uint16)
            >>> puma.render_volume(ws, style='edges')
        """
        if isinstance(nparray, np.ndarray):
            if nparray.ndim == 3:
                self.matrix = nparray.copy().astype(np.uint16)
            else:
                raise Exception("Wrong nparray ndim, 3 dimensions required. Leaving matrix unchanged")
        else:
            print_warning("Wrong nparray type. Leaving matrix unchanged")

    def set_orientation(self, nparray):
        """ Set orientation with numpy array

            :param nparray: array of shape (X,Y,Z, 3) to be assigned to the orientation variable
            :type nparray: np.ndarray
            :return: None

            :Example:
            >>> import pumapy as puma
            >>> ws = puma.Workspace()
            >>> ws.set_orientation(np.random.rand(10, 10, 10, 3))
            >>> puma.render_orientation(ws, solid_color=None)
        """
        if isinstance(nparray, np.ndarray):
            if nparray.ndim == 4 and nparray.shape[3] == 3:
                self.orientation = nparray.copy().astype(float)
            else:
                raise Exception("Wrong nparray ndim, 4 dimensions required as (x,y,z,3). Leaving orientation unchanged")
        else:
            raise Exception("Wrong nparray type. Leaving orientation unchanged")

    def copy(self):
        """ Create a copy of the workspace

            :return: copy of workspace
            :rtype: pumapy.Workspace
        """
        return deepcopy(self)

    def __getitem__(self, key):
        return self.matrix[key]

    def __setitem__(self, key, value):
        self.matrix[key] = value

    def get_size(self):
        """ Return size

            :return: number of voxels
            :rtype: int
        """
        return self.matrix.size

    def len_x(self):
        """ Return x dimension length

            :return: number of voxels in x dimension
            :rtype: int
        """
        return self.matrix.shape[0]

    def len_y(self):
        """ Return y dimension length

            :return: number of voxels in y dimension
            :rtype: int
        """
        return self.matrix.shape[1]

    def len_z(self):
        """ Return z dimension length

            :return: number of voxels in z dimension
            :rtype: int
        """
        return self.matrix.shape[2]

    def get_shape(self):
        """ Return shape of domain

            :return: shape of domain (matrix)
            :rtype: (int, int, int)
        """
        return self.matrix.shape

    def min(self):
        """ Return minimum of domain

            :return: minimum
            :rtype: int
        """
        return np.min(self.matrix)

    def max(self):
        """ Return maximum of domain

            :return: maximum
            :rtype: int
        """
        return np.max(self.matrix)

    def average(self):
        """ Return average of domain

            :return: grayscale average
            :rtype: float
        """
        return self.matrix.mean()

    def unique_values(self):
        """ Return unique values in domain matrix variable

            :return: unique values
            :rtype: np.ndarray
        """
        return np.unique(self.matrix)

    def unique_values_counts(self):
        """ Return unique values and their counts in domain matrix variable

            :return: unique values and counts
            :rtype: (np.ndarray, int)
        """
        return np.unique(self.matrix, return_counts=True)

    def orientation_magnitude(self):
        """ Return orientation vector's magnitude

            :return: orientation field magnitude
            :rtype: np.ndarray
        """
        return np.linalg.norm(self.orientation, axis=3)

    def resize_new_matrix(self, shape, value=None):
        """ Resize matrix numpy array

            :param shape: shape of workspace to be resized
            :type shape: (int, int, int)
            :param value: value to be assigned to the new resized matrix variable
            :type value: int

            :Example:
            >>> import pumapy as puma
            >>> ws = puma.Workspace()
            >>> ws.resize_new_matrix((10, 10, 10), value=3)
            >>> puma.render_volume(ws, style='edges')
        """
        if isinstance(shape, tuple) and len(shape) == 3:
            if value is None:
                self.matrix = np.zeros(shape, dtype=np.uint16)
            else:
                self.matrix = np.full(shape, value, dtype=np.uint16)
        else:
            raise Exception("Wrong shape, tuple with 3 dimensions required.")

    def resize_new_orientation(self, shape, orientation_value=None):
        """ Resize orientation numpy array

            :param shape: shape of workspace to be resized
            :type shape: (int, int, int)
            :param orientation_value: vector to be assigned to the new resized orientation variable
            :type orientation_value: (float, float, float)

            :Example:
            >>> import pumapy as puma
            >>> ws = puma.Workspace()
            >>> ws.resize_new_orientation((10, 10, 10), orientation_value=(1., 0., 0.))
            >>> puma.render_orientation(ws)
        """
        if isinstance(shape, tuple) and len(shape) == 3:
            if orientation_value is None:
                self.orientation = np.zeros(list(shape) + [3], dtype=float)
            else:
                if isinstance(orientation_value, tuple) and len(orientation_value) == 3:
                    self.orientation = np.tile(orientation_value, list(shape) + [1]).astype(float)
                else:
                    raise Exception("vectorvalue has to be (float, float, float).")
        else:
            raise Exception("Wrong shape, tuple with 3 dimensions required.")

    def create_orientation(self):
        """ Create orientation field of the same size as the matrix 

            :return: None

            :Example:
            >>> import pumapy as puma
            >>> ws = puma.Workspace.from_shape((10, 10, 10))
            >>> ws.create_orientation()
            >>> print(ws.orientation.shape)
        """
        self.orientation = np.zeros(list(self.matrix.shape) + [3], dtype=float)

    def resize(self, shape, segmented, anti_aliasing=True, interpolation_order=1):
        """ Resize both matrix and orientation (if present) by rescaling the content to specified size

            :param shape: shape of workspace to be resized
            :type shape: (int, int, int)
            :param segmented: specifying whether the domain is already segmented (True) or grayscales (False)
            :type segmented: bool
            :param anti_aliasing: if aliasing is to be prevented applying a Gaussian filter to smooth
                before scaling. If domain is segmented, automatically set to False in order to preserve domain
            :type anti_aliasing: bool
            :param interpolation_order: order of the interpolation spline used.
                For segmented, it is enforced to be 0,which is 'nearest neighbor' to preserve the segmentation
            :type interpolation_order: int
            :return: None

            :Example:
            >>> import pumapy as puma
            >>> ws_fiberform = puma.import_3Dtiff(puma.path_to_example_file("200_fiberform.tif"), 1.3e-6)
            >>> ws2 = ws_fiberform.copy()
            >>> ws2.resize((100, 200, 200), segmented=False)
            >>> puma.compare_slices(ws_fiberform, ws2)
        """
        if isinstance(shape, tuple) and len(shape) == 3:
            if self.orientation.shape[:3] == self.matrix.shape:
                self.orientation = trans.resize(self.orientation, list(shape) + [3], order=0,
                                                preserve_range=True, anti_aliasing=False)
            if segmented:
                self.matrix = trans.resize(self.matrix, shape, order=0, anti_aliasing=False, preserve_range=True)
            else:
                self.matrix = trans.resize(self.matrix, shape, order=interpolation_order,
                                           anti_aliasing=anti_aliasing, preserve_range=True)
            self.matrix = self.matrix.astype('uint16')
        else:
            raise Exception("Wrong shape, tuple with 3 dimensions required.")

    def rescale(self, scale, segmented, anti_aliasing=True, interpolation_order=1):
        """ Rescale both matrix and orientation (if present) by rescaling the content by a specified factor

            :param scale: specifying the scaling factor
            :type scale: float
            :param segmented: specifying whether the domain is already segmented (True) or grayscales (False)
            :type segmented: bool
            :param anti_aliasing: if aliasing is to be prevented applying a Gaussian filter
                to smooth before scaling. If domain is segmented, automatically set to False in order to preserve domain
            :type anti_aliasing: bool
            :param interpolation_order: order of the interpolation spline used.
                For segmented, it is enforced to be 0, which is 'nearest neighbor' to preserve the segmentation
            :type interpolation_order: int
            :return: None

            :Example:
            >>> import pumapy as puma
            >>> ws_fiberform = puma.import_3Dtiff(puma.path_to_example_file("200_fiberform.tif"), 1.3e-6)
            >>> ws2 = ws_fiberform.copy()
            >>> ws2.rescale(0.5, segmented=False)
            >>> puma.compare_slices(ws_fiberform, ws2)  # pay attention to the shape
        """

        unit_dim_check = None
        if min(self.get_shape()) == 1:
            if self.len_x() == 1:
                unit_dim_check = 0
            elif self.len_y() == 1:
                unit_dim_check = 1
            elif self.len_z() == 1:
                unit_dim_check = 2
            self.matrix = np.squeeze(self.matrix)

        if self.orientation.shape[:3] == self.matrix.shape:
            self.orientation = trans.rescale(self.orientation, scale, order=0, multichannel=True,
                                             preserve_range=True, anti_aliasing=False)
        if segmented:
            self.matrix = trans.rescale(self.matrix, scale, order=0, anti_aliasing=False, preserve_range=True)
        else:
            self.matrix = trans.rescale(self.matrix, scale, order=interpolation_order,
                                        anti_aliasing=anti_aliasing, preserve_range=True)

        if unit_dim_check is not None:
            self.matrix = np.expand_dims(self.matrix, axis=unit_dim_check)

        self.matrix = self.matrix.astype('uint16')
        self.voxel_length /= scale
        print("Rescaled workspace size: {}".format(self.get_shape()))

    def set(self, matrix_value=None, orientation_value=None):
        """ Set all elements in matrix equal to value (and orientation to vectorvalue is passed)

            :param matrix_value: value to fill to the matrix variable. NB this value will be turned into np.uint16
            :type matrix_value: int
            :param orientation_value: vector to fill to the orientation variable
            :type orientation_value: ((float, float, float))
            :return: None

            :Example:
            >>> import pumapy as puma
            >>> ws = puma.Workspace.from_shape_value((10, 10, 10), 5)
            >>> print(ws.unique_values())
            >>> ws.set(matrix_value=4)
            >>> print(ws.unique_values())
        """
        check = True
        if matrix_value is not None:
            if np.issubdtype(type(matrix_value), np.integer) and 0 <= matrix_value <= 65535:
                self.matrix.fill(np.uint16(matrix_value))
                check = False
            else:
                raise Exception("matrix_value has to be np.uint16 (i.e. 0 <= matrix_value <= 65535).")
        if orientation_value is not None:
            if isinstance(orientation_value, tuple) and len(orientation_value) == 3:
                if self.orientation.shape[:3] != self.matrix.shape:
                    self.resize_new_orientation(self.matrix.shape, orientation_value=orientation_value)
                else:
                    self.orientation[:] = orientation_value
                check = False
            else:
                raise Exception("orientation_value has to be (float, float, float).")
        if check:
            print_warning("No changes have been made to the Workspace.")

    def apply_mask(self, mask, apply_to_orientation=False):
        """ Apply mask of same size as the matrix by leaving the mask's 1s unchanged and setting mask's 0s to 0

            :param mask: array of type bool with same size as matrix
            :type mask: np.ndarray
            :param apply_to_orientation: specifying whether the mask is to be applied to the orientation (if present)
            :type apply_to_orientation: bool
            :return: None

            :Example:
            >>> import pumapy as puma
            >>> ws = puma.Workspace.from_shape_value_vector((10, 10, 10), 1, (0.4, 2, 5))
            >>> mask = np.random.randint(255, size=(10, 10, 10)) > 100
            >>> ws.apply_mask(mask, apply_to_orientation=True)
            >>> puma.render_volume(ws, style='edges')
        """
        if isinstance(mask, np.ndarray) and mask.dtype == 'bool':
            if mask.shape == self.matrix.shape:
                self.matrix[~mask] = 0
                if self.orientation.shape[:3] == self.matrix.shape and apply_to_orientation:
                    self.orientation[~mask] = 0
            else:
                raise Exception("The mask has to be of the same size as the Workspace matrix.")
        else:
            raise Exception("The mask has to be a Numpy array of type bool.")

    def set_material_id(self, cutoff, value):
        """ Threshold the workspace according to cutoff (i.e. tuple with inclusive range to set)

            :param cutoff: convert a range of grayscale values specified by the cutoff into an single ID number
            :type cutoff: (int, int)
            :param value: ID number to assign to range
            :type value: int
            :return: None

            :Example:
            >>> import pumapy as puma
            >>> ws_fiberform = puma.import_3Dtiff(puma.path_to_example_file("200_fiberform.tif"), 1.3e-6)
            >>> ws2 = ws_fiberform.copy()
            >>> ws2.set_material_id((0, 100), 0)    # NB the order of these operations is important!
            >>> ws2.set_material_id((100, 255), 1)  # this is why binarize and binarize_range should be preferred
            >>> puma.compare_slices(ws_fiberform, ws2)
        """
        if value < 0:
            raise Exception("Value ID can only be positive. Leaving matrix unchanged.")
        if value > 1000000:
            raise Exception("Value ID cannot be > 1000000. Leaving matrix unchanged.")

        self.matrix = np.where(np.logical_and(self.matrix >= cutoff[0], self.matrix <= cutoff[1]), np.uint16(value), self.matrix)

        self.log.log_section("Set Material ID")
        self.log.log_line(str(cutoff) + " -> " + str(value))
        self.log.write_log()

    def binarize(self, threshold):
        """ Binarize the workspace according to threshold, inclusive for higher range set to 1, lower to 0

            :param threshold: grayscale value dividing the domain into 0s and 1s (threshold turns into 1)
            :type threshold: int
            :return: None

            :Example:
            >>> import pumapy as puma
            >>> ws_fiberform = puma.import_3Dtiff(puma.path_to_example_file("200_fiberform.tif"), 1.3e-6)
            >>> ws2 = ws_fiberform.copy()
            >>> ws2.binarize(100)
            >>> puma.compare_slices(ws_fiberform, ws2)
        """
        self.matrix = np.where(self.matrix < threshold, np.uint16(0), np.uint16(1))

        self.log.log_section("Binarize domain")
        self.log.log_line(">" + str(threshold) + " -> 1; 0 otherwise")
        self.log.write_log()

    def binarize_range(self, ones_cutoff):
        """ Binarize the workspace according to range within cutoff, inclusive for cutoff ints set to 1, rest to 0

            :param ones_cutoff: convert a range of grayscale values specified by the cutoff into 1s, rest into 0s
            :type ones_cutoff: (int, int)
            :return: None

            :Example:
            >>> import pumapy as puma
            >>> ws_fiberform = puma.import_3Dtiff(puma.path_to_example_file("200_fiberform.tif"), 1.3e-6)
            >>> ws2 = ws_fiberform.copy()
            >>> ws2.binarize_range((100, 255))
            >>> puma.compare_slices(ws_fiberform, ws2)
        """
        self.matrix = np.where(np.logical_and(self.matrix >= ones_cutoff[0], self.matrix <= ones_cutoff[1]), np.uint16(1), np.uint16(0))

        self.log.log_section("Binarize domain")
        self.log.log_line(str(ones_cutoff[0]) + " < matrix < " + str(ones_cutoff[1]) + " -> 1; 0 otherwise")
        self.log.write_log()

    def rotate(self, degrees, around_axis, reshape=True, boundary_mode='reflect', apply_to_orientation=True):
        """ Rotate domain by specified degrees

            :param degrees: degrees to rotate domain
            :type degrees: float
            :param around_axis: specify around what axis to perform the rotation. It can be 'x', 'y' or 'z'
            :type around_axis: string
            :param reshape: specify whether to reshape the domain (and therefore contain every voxel - reshape=True)
                            or keep its original size (reshape=False)
            :type reshape: bool
            :param boundary_mode: specifying what to do with the boundaries.
                Options: ‘reflect’, ‘constant’, ‘nearest’, ‘mirror’, ‘wrap’
            :type boundary_mode: string
            :param apply_to_orientation: specify whether to apply rotation to the orientation, if present
            :type apply_to_orientation: bool
            :return: None

            :Example:
            >>> import pumapy as puma
            >>> ws_fiberform = puma.import_3Dtiff(puma.path_to_example_file("200_fiberform.tif"), 1.3e-6)
            >>> ws_copy = ws_fiberform.copy()
            >>> ws_copy.rotate(45, 'z', reshape=False, boundary_mode='reflect')
            >>> puma.compare_slices(ws_fiberform, ws_copy)
            >>> ws_copy = ws_fiberform.copy()
            >>> ws_copy.rotate(45, 'z', reshape=True, boundary_mode='constant')
            >>> puma.compare_slices(ws_fiberform, ws_copy)
            >>> ws_copy = ws_fiberform.copy()
            >>> ws_copy.rotate(45, 'z', reshape=True, boundary_mode='reflect')
            >>> puma.compare_slices(ws_fiberform, ws_copy)
        """
        if around_axis == 'x':
            axes = (1, 2)
        elif around_axis == 'y':
            axes = (0, 2)
        elif around_axis == 'z':
            axes = (0, 1)
        else:
            raise Exception("Axis not recognized: around_axis can only be 'x', 'y' or 'z'")

        self.matrix = rotate(self.matrix, angle=degrees, axes=axes, mode=boundary_mode, reshape=reshape)

        if self.orientation.shape[:3] == self.matrix.shape and apply_to_orientation:
            from scipy.spatial.transform import Rotation
            rotation_degrees = degrees
            rotation_radians = np.radians(rotation_degrees)
            if around_axis == 'x':
                rotation_axis = np.array([1, 0, 0])
            elif around_axis == 'y':
                rotation_axis = np.array([0, 1, 0])
            else:
                rotation_axis = np.array([0, 0, 1])
            rotation_vector = rotation_radians * rotation_axis
            rotation = Rotation.from_rotvec(rotation_vector)
            for i in range(self.len_x()):
                for j in range(self.len_y()):
                    for k in range(self.len_z()):
                        self.orientation[i, j, k] = rotation.apply(self.orientation[i, j, k])

            self.orientation = rotate(self.orientation, angle=degrees, axes=axes, mode=boundary_mode,
                                      reshape=reshape, order=0)

    def porosity(self, cutoff=(0, 0)):
        """ Compute porosity of domain

            :param cutoff: void cutoff
            :type cutoff: (int, int)
            :return: volume fraction
            :rtype: float

            :Example:
            >>> import pumapy as puma
            >>> ws = puma.import_3Dtiff(puma.path_to_example_file("100_fiberform.tif"), 1.3e-6)
            >>> ws.porosity(cutoff=(0, 89)))
        """
        return pumapy.compute_volume_fraction(self, cutoff)


    def show_matrix(self):
        """ Print content of matrix domain's variable """
        if isinstance(self, Workspace):
            x, y, z = self.matrix.shape
        elif isinstance(self, np.ndarray):
            x, y, z = self.shape
        else:
            raise Exception("Print can only be called on a Workspace or Numpy array.")
        print()
        print("3D Workspace:")
        # Printing coordinate system used
        print("  o---> y")
        print("  |")
        print("x v")
        print('[', end='')
        for k in range(z):
            print("(:,:,{})".format(k))
            print('[', end='')
            for i in range(x):
                print('[', end='')
                for j in range(y):
                    print(self[i, j, k], end='')
                    if j != y - 1:
                        print(' ', end='')
                print(']', end='')
                if i != x - 1:
                    print()
            if k != z - 1:
                print()
                print()
        print(']')

    def show_orientation(self, dec=1):
        """ Print content of orientation domain's variable """
        if isinstance(self, Workspace):
            dims = self.matrix.shape
        elif isinstance(self, np.ndarray):
            if self.shape[3] == 3:
                dims = self.shape
            else:
                raise Exception("Numpy array has to be of size (x,y,z,3,...).")
        else:
            raise Exception("Print can only be called on a Workspace or Numpy array.")
        print()
        print("3D Orientation:")
        # Printing coordinate system used
        print("  o---> y")
        print("  |")
        print("x v")
        print('[', end='')
        if isinstance(self, Workspace):
            for k in range(dims[2]):
                print("(:,:,{})".format(k))
                print('[', end='')
                for i in range(dims[0]):
                    print('[', end='')
                    for j in range(dims[1]):
                        print('({:.{}f}, {:.{}f}, {:.{}f})'.format(self.orientation[i, j, k, 0], dec,
                                                                   self.orientation[i, j, k, 1], dec,
                                                                   self.orientation[i, j, k, 2], dec), end='')
                        if j != dims[1] - 1:
                            print(' ', end='')
                    print(']', end='')
                    if i != dims[0] - 1:
                        print()
                if k != dims[2] - 1:
                    print()
                    print()
        else:
            for k in range(dims[2]):
                print("(:,:,{})".format(k))
                print('[', end='')
                for i in range(dims[0]):
                    print('[', end='')
                    for j in range(dims[1]):
                        print('({}, {}, {})'.format(self[i, j, k, 0],
                                                    self[i, j, k, 1],
                                                    self[i, j, k, 2]), end='')
                        if j != dims[1] - 1:
                            print(' ', end='')
                    print(']', end='')
                    if i != dims[0] - 1:
                        print()
                if k != dims[2] - 1:
                    print()
                    print()
        print(']')



import unittest
import pumapy as puma
import numpy as np


class TestFilters(unittest.TestCase):
    
    def setUp(self):
        self.ws = puma.import_3Dtiff(puma.path_to_example_file("100_fiberform.tif"), 1.3e-6)
        self.shape = self.ws.get_shape()
        
    def test_median(self):
        puma.filter_median(self.ws, size=10)
        
        np.testing.assert_equal(self.ws.matrix.shape, self.shape)

    def test_gaussian(self):
        puma.compute_orientation_st(self.ws, cutoff=(90, 255))
        puma.filter_gaussian(self.ws, sigma=2, apply_on_orientation=True)
        
        np.testing.assert_equal(self.ws.matrix.shape, self.shape)
    
    def test_edt(self):
        puma.filter_edt(self.ws, cutoff=(90, 255))
        
        np.testing.assert_equal(self.ws.matrix.shape, self.shape)
    
    def test_mean(self):
        puma.filter_mean(self.ws, size=10)

        np.testing.assert_equal(self.ws.matrix.shape, self.shape)
    
    def test_erode(self):
        puma.filter_erode(self.ws, (90, 255), size=6)

        np.testing.assert_equal(self.ws.matrix.shape, self.shape)
    
    def test_dilate(self):
        puma.filter_dilate(self.ws, cutoff=(90, 255), size=5)

        np.testing.assert_equal(self.ws.matrix.shape, self.shape)
    
    def test_opening(self):
        puma.filter_opening(self.ws, cutoff=(90, 255), size=3)

        np.testing.assert_equal(self.ws.matrix.shape, self.shape)
    
    def test_closing(self):
        puma.filter_closing(self.ws, cutoff=(90, 255), size=3)

        np.testing.assert_equal(self.ws.matrix.shape, self.shape)
    

if __name__ == '__main__':
    unittest.main()



from pumapy.physicsmodels.fe_permeability import Permeability


def compute_permeability(workspace, solid_cutoff, tol=1e-8, maxiter=10000, solver_type='minres', display_iter=True):
    """ Compute the permeability using first order Q1-Q1 Finite Element solver and periodic BC on the sides

        :param workspace: domain
        :type workspace: pumapy.Workspace
        :param solid_cutoff: specify the solid phase
        :type solid_cutoff: (int, int)
        :param tol: tolerance for iterative solver
        :type tol: float
        :param maxiter: maximum Iterations for solver
        :type maxiter: int
        :param solver_type: solver type, options: 'minres' (default), 'direct', 'cg', 'bicgstab'
        :type solver_type: string
        :param display_iter: display iteration in iterative solver
        :type display_iter: bool
        :return: effective permeability (3x3 matrix), velocity and pressure fields for x, y, z directions (u_x, p_x, etc)
        :rtype: numpy.ndarray, numpy.ndarray, numpy.ndarray, numpy.ndarray, numpy.ndarray, numpy.ndarray, numpy.ndarray

        :Example:
        >>> import pumapy as puma
        >>> ws = puma.generate_random_fibers(shape=(100, 100, 100), radius=3, porosity=0.7, phi=90, theta=90, length=200)
        >>> keff, u_x, p_x, u_y, p_y, u_z, p_z = puma.compute_permeability(ws, (1, ws.max()))
        >>> puma.render_orientation(u_x, scale_factor=5e11, solid_color=None)
    """
    solver = Permeability(workspace, solid_cutoff, tol, maxiter, solver_type, display_iter)

    solver.error_check()

    solver.log_input()
    solver.compute()
    solver.log_output()
    return solver.keff, solver.u_x, solver.p_x, solver.u_y, solver.p_y, solver.u_z, solver.p_z
